# kate: default-dictionary en_US

## This file is part of the 'stringi' package for R.
## Copyright (c) 2013-2020, Marek Gagolewski <https://www.gagolewski.com>
## All rights reserved.
##
## Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
## 1. Redistributions of source code must retain the above copyright notice,
## this list of conditions and the following disclaimer.
##
## 2. Redistributions in binary form must reproduce the above copyright notice,
## this list of conditions and the following disclaimer in the documentation
## and/or other materials provided with the distribution.
##
## 3. Neither the name of the copyright holder nor the names of its
## contributors may be used to endorse or promote products derived from
## this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
## FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
## OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
## EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


invisible(NULL)  # TO DO

# #' @title
# #' t.b.d.
# #'
# #' @description
# #' t.b.d.
# #'
# #' @details
# #' t.b.d.
# #'
# #' @param x numeric vector
# #' @param ruleset single string; one of \code{spellout}, \code{ordinal}, \code{duration},
# #'                 \code{numbering_system}?
# #' @param locale \code{NULL} or \code{''} for case mapping following
# #' the conventions of the default locale, or a single string with
# #' locale identifier, see \link{stringi-locale}.
# #'
# #' @return
# #' Returns a character vector
# #'
# #' @family number_format
# #' @export
# stri_num_rbnf_format <- function(x, ruleset='spellout', locale=NULL) # stri_format_rbnf?
# {
#    stop('t.b.d.')
# }
#
#
# #' @title
# #' t.b.d.
# #'
# #' @description
# #' t.b.d.
# #'
# #' @details
# #' t.b.d.
# #'
# #' @param str character vector
# #' @param ruleset single string; one of \code{spellout}, \code{ordinal}, \code{duration},
# #'                 \code{numbering_system}?
# #' @param locale \code{NULL} or \code{''} for case mapping following
# #' the conventions of the default locale, or a single string with
# #' locale identifier, see \link{stringi-locale}.
# #' @param lenient single logical value; ....
# #'
# #' @return
# #' Returns a numeric vector (list of numeric vectors?)
# #'
# #' @family number_format
# #' @export
# stri_num_rbnf_parse <- function(str, ruleset='spellout', locale=NULL, lenient=TRUE) # stri_parse_rbnf?
# {
#    stop('t.b.d.')
# }
